@@ -28,6 +28,7 @@
 #include <sys/ioctl.h>
 #include <linux/vzcalluser.h>
 #include <sys/personality.h>
+#include <linux/reboot.h>
 
 #include "vzerror.h"
 #include "res.h"
@@ -83,11 +84,15 @@ int vz_env_create_data_ioctl(vps_handler *h,
 			sleep(1);
 		errcode = ioctl(h->vzfd, VZCTL_ENV_CREATE_DATA, data);
 	} while (errcode < 0 && errno == EBUSY && retry++ < ENVRETRY);
+
+	if (errcode >= 0) {
+		/* Clear supplementary group IDs */
+		setgroups(0, NULL);
 #ifdef  __x86_64__
-	/* Set personality PER_LINUX32 for i386 based VEs */
-	if (errcode >= 0)
+		/* Set personality PER_LINUX32 for i386 based VEs */
 		set_personality32();
 #endif
+	}
 	return errcode;
 }
 
@@ -105,11 +110,14 @@ int vz_env_create_ioctl(vps_handler *h, envid_t veid, int flags)
 			sleep(1);
 		errcode = ioctl(h->vzfd, VZCTL_ENV_CREATE, &env_create);
 	} while (errcode < 0 && errno == EBUSY && retry++ < ENVRETRY);
+	if (errcode >= 0 && (flags & VE_ENTER)) {
+		/* Clear supplementary group IDs */
+		setgroups(0, NULL);
 #ifdef  __x86_64__
-	/* Set personality PER_LINUX32 for i386 based VEs */
-	if (errcode >= 0 && (flags & VE_ENTER))
+		/* Set personality PER_LINUX32 for i386 based VEs */
 		set_personality32();
 #endif
+	}
 	return errcode;
 }
 
@@ -153,7 +161,7 @@ vps_handler *vz_open(envid_t veid)
 		goto err;
 	}
 	h = calloc(1, sizeof(*h));
-	if (h == NULL) 
+	if (h == NULL)
 		goto err;
 	h->vzfd = vzfd;
 	h->stdfd = stdfd;
@@ -225,7 +233,7 @@ int vz_chroot(char *root)
 	struct sigaction act;
 
 	if (root == NULL) {
-		logger(-1, 0, "vz_chroot: VE root is not specified");	
+		logger(-1, 0, "vz_chroot: VE root is not specified");
 		return VZ_VE_ROOT_NOTSET;
 	}
 	if (chdir(root)) {
@@ -384,6 +392,11 @@ try:
 	*/
 	if (read(wait_p, &ret, sizeof(ret)) != 0)
 		return 0;
+	if ((fd = open("/dev/null", O_RDWR)) != -1) {
+		dup2(fd, 0);
+		dup2(fd, 1);
+		dup2(fd, 2);
+	}
 	logger(10, 0, "Starting init");
 	execve("/sbin/init", argv, envp);
 	execve("/etc/init", argv, envp);
@@ -561,7 +574,7 @@ int vps_start_custom(vps_handler *h, envid_t veid, vps_param *param,
 	}
 	if (!vps_is_mounted(res->fs.root)) {
 		/* increase quota to perform setup */
-		quouta_inc(&res->dq, 100);	
+		quouta_inc(&res->dq, 100);
 		if ((ret = vps_mount(h, veid, &res->fs, &res->dq, skip)))
 			return ret;
 		quouta_inc(&res->dq, -100);
@@ -591,7 +604,7 @@ int vps_start_custom(vps_handler *h, envid_t veid, vps_param *param,
 	}
 	if (!(skip & SKIP_ACTION_SCRIPT)) {
 		snprintf(buf, sizeof(buf), VPS_CONF_DIR "%d.%s", veid,
-			START_PREFIX);	
+			START_PREFIX);
 		if (stat_file(buf)) {
 			if (vps_exec_script(h, veid, res->fs.root, NULL, NULL,
 				buf, NULL, 0))
@@ -656,10 +669,9 @@ err:
  *
  * @param h		VE handler.
  * @param veid		VE id.
- * @param res		VE resourses.
- * @param d_actions	distribution specific actions.
+ * @param param	VE parameters.
  * @param skip		flags to skip VE setup (SKIP_SETUP|SKIP_ACTION_SCRIPT)
- * @param action	modules list, used to call setup() callback
+ * @param mod		modules list, used to call setup() callback
  * @return		0 on success.
  */
 int vps_start(vps_handler *h, envid_t veid, vps_param *param,
@@ -700,9 +712,9 @@ static int real_env_stop(vps_handler *h, envid_t veid, char *vps_root,
 		}
 		case M_KILL:
 		{
-			kill(-1, SIGTERM);
-			sleep(1);
-			kill(1, SIGKILL);
+			syscall(__NR_reboot, LINUX_REBOOT_MAGIC1,
+					LINUX_REBOOT_MAGIC2,
+					LINUX_REBOOT_CMD_POWER_OFF, NULL);
 			break;
 		}
 	}
@@ -744,12 +756,12 @@ kill_vps:
 		ret = VZ_RESOURCE_ERROR;
 		logger(-1, errno, "Can not fork");
 		goto err;
-	
+
 	} else if (pid == 0) {
 		ret = real_env_stop(h, veid, root, M_KILL);
 		exit(ret);
 	}
-	ret = VZ_STOP_ERROR;	
+	ret = VZ_STOP_ERROR;
 	for (i = 0; i < MAX_SHTD_TM; i++) {
 		usleep(500000);
 		if (!vps_is_run(h, veid)) {
@@ -771,7 +783,7 @@ err:
  *
  * @param h		VE handler.
  * @param veid		VE id.
- * @param res		VE resourses.
+ * @param param	VE parameters.
  * @param stop_mode	stop mode, one of (M_REBOOT M_HALT M_KILL).
  * @param skip		flag to skip run action script (SKIP_ACTION_SCRIPT)
  * @param action	modules list, used to call cleanup() callback.
@@ -792,7 +804,7 @@ int vps_stop(vps_handler *h, envid_t veid, vps_param *param, int stop_mode,
 	}
 	if (!(skip & SKIP_ACTION_SCRIPT)) {
 		snprintf(buf, sizeof(buf), VPS_CONF_DIR "%d.%s", veid,
-			STOP_PREFIX);	
+			STOP_PREFIX);
 		if (stat_file(buf)) {
 			if (vps_exec_script(h, veid, res->fs.root, NULL, NULL,
 				buf, NULL, 0))
@@ -816,6 +828,13 @@ end:
 	return ret;
 }
 
+/** Restart VE.
+ *
+ * @param h		VE handler.
+ * @param veid		VE id.
+ * @param param	VE parameters.
+ * @return		0 on success.
+ */
 int vps_restart(vps_handler *h, envid_t veid, vps_param *param)
 {
 	int ret;
@@ -826,6 +845,11 @@ int vps_restart(vps_handler *h, envid_t veid, vps_param *param)
 	{
 		return ret;
 	}
+	if (param->opt.start_disabled == YES)
+	{
+		logger(-1, 0, "VE start disabled");
+		return VZ_VE_START_DISABLED;
+	}
 	ret = vps_start(h, veid, param, SKIP_NONE, NULL);
 
 	return ret;
